@@ -212,20 +212,76 @@ def v_text_clparameter(self, node):
 
 def specializer(sexp, state):
     result = StringIO()
+    result.write("(")
     for atom in sexp:
         if atom.startswith("KEYWORD:"):
             result.write("(EQL :%s)" % atom.split(":")[-1])
         else:
-            result.write(":cl:symbol:`~%s`" % atom)
+            result.write(atom)
         result.write(" ")
-
-    node = nodes.list_item()
+    result.write(")")
+    node = nodes.inline()
     result.seek(0)
-    lines = string2lines(result.read())
+
+    lines = string2lines(result.read().lower())
     state.nested_parse(StringList(lines), 0, node)
     return node
 
 
+def specializer_xref(symbol, sexp, state):
+    result = StringIO()
+    first = True
+    for atom in sexp:
+        if first:
+            first = False
+        else:
+            result.write(" ")
+
+        if atom.startswith("KEYWORD:"):
+            result.write("(EQL :%s)" % atom.split(":")[-1])
+        else:
+            result.write(atom)
+
+    target = " ".join([a.lower() for a in sexp])
+    node = nodes.list_item()
+    result.seek(0)
+    xref = ":cl:method:`(%s) <%s %s>`" % (result.read().lower(), symbol, target)
+
+    lines = string2lines(xref)
+    state.nested_parse(StringList(lines), 0, node)
+    return node
+
+
+def qualify_sexp(package, sexp):
+    """if the sexp contains atoms that don't have a package
+    then qualify them."""
+    sexp_ret = []
+    for atom in sexp:
+        if atom.startswith(":"):
+            sexp_ret.append("keyword" + atom)
+        elif ":" in atom:
+            sexp_ret.append(atom)
+        else:
+            sexp_ret.append(package + ":" + atom)
+    return sexp_ret
+
+
+def fieldlist_index(node):
+    """find the index of a field list in a content node."""
+    for i, n in enumerate(node):
+        if isinstance(n, nodes.field_list):
+            return i
+
+
+def get_content_node(node):
+    """Search through and find the content node from a signature."""
+    for subnode in node:
+        if isinstance(subnode, addnodes.desc):
+            for subsubnode in subnode:
+                if isinstance(subsubnode, addnodes.desc_content):
+                    return subsubnode
+
+
 class SpecializerField(Field):
     """
     """
@@ -283,7 +339,7 @@ class SEXP(object):
         if token.lower() in lambda_list_keywords:
             param["lambda_keyword"] = True
         if token.startswith(":"):
-            praam["keyword"] = True
+            param["keyword"] = True
         signode.append(param)
 
 
@@ -304,12 +360,10 @@ class CLsExp(ObjectDescription):
 
     def handle_signature(self, sig, signode):
         symbol_name = []
-        type = []
         package = self.env.temp_data.get('cl:package')
         objtype = self.get_signature_prefix(sig)
         sig_split = sig.split(" ")
         sig = sig_split[0]
-        args = sig_split[1:]
         signode.append(addnodes.desc_annotation(objtype, objtype))
         lisp_args = ARGS[package].get(sig.upper(), "")
 
@@ -337,48 +391,66 @@ class CLsExp(ObjectDescription):
         if not symbol_name:
             raise Exception("Unknown symbol type for signature %s" % sig)
         record_use(package, symbol_name, self.objtype)
-        return objtype.strip(), symbol_name.upper()
+        return objtype.strip(), symbol_name
+
+    def get_field_list(self, node):
+        """Return the node's field list, if there isn't one then
+        create it first."""
+        # Add a field list if there isn't one
+        if not node[1][1].children:
+            node[1][1].append(nodes.field_list())
+        if not isinstance(node[1][1][0], nodes.field_list):
+            node[1][1].children.insert(0, nodes.field_list())
+        return node[1][1][0]
 
     def get_index_text(self, name, type):
         return _('%s (Lisp %s)') % (name.lower().split(":")[-1], type)
 
+    def get_index_name(self, name, type):
+        return type + ":" + name
+
     def get_signature_prefix(self, sig):
         return self.objtype + ' '
 
+    def cl_symbol_name(self):
+        return self.names[0][1].upper()
+
     def add_target_and_index(self, name, sig, signode):
-        # note target
+        # node target
         type, name = name
 
         if 'cl:package' in self.env.temp_data:
             package = self.options.get(
                 'module', self.env.temp_data.get('cl:package'))
-            name = package + ":" + name
+            name = package.lower() + ":" + name
         else:
-            package = ""
+            return
 
-        if type != "method" and name not in self.state.document.ids:
+        indexname = self.get_index_name(name, type)
+        if name not in self.state.document.ids:
             signode['names'].append(name)
-            signode['ids'].append(name)
+            signode['ids'].append(indexname)
             signode['first'] = (not self.names)
             self.state.document.note_explicit_target(signode)
             inv = self.env.domaindata['cl']['symbols']
+            # TODO (RS) reenable this checking based on doc and type.
+            # if name in inv:
+            #     self.state_machine.reporter.warning(
+            #         'duplicate symbol description of %s, ' % name +
+            #         'other instance in ' + self.env.doc2path(inv[name][0]),
+            #         line=self.lineno)
             if name in inv:
-                self.state_machine.reporter.warning(
-                    'duplicate symbol description of %s, ' % name +
-                    'other instance in ' + self.env.doc2path(inv[name][0]),
-                    line=self.lineno)
-            inv[name] = (self.env.docname, self.objtype)
+                inv[name].append((self.env.docname, self.objtype))
+            else:
+                inv[name] = [(self.env.docname, self.objtype)]
 
-        if not type == "method":
-            indextext = self.get_index_text(name, type)
-            if indextext:
-                self.indexnode['entries'].append(('single', indextext, name, ''))
+        indextext = self.get_index_text(name, type)
+        if indextext:
+            self.indexnode['entries'].append(('single', indextext, indexname, ''))
 
     def run_add_doc(self, result):
         package = self.env.temp_data.get('cl:package')
         name = self.names[0][1]
-        description = result[1][-1]
-        node = addnodes.desc_content()
         if not package:
             self.state_machine.reporter.warning("No package specified for symbol %s." %
                                                 name)
@@ -391,46 +463,27 @@ class CLsExp(ObjectDescription):
                                                 (package, name))
         if not string:
             return
+        node = addnodes.desc_content()
         lines = string2lines(string)
         self.state.nested_parse(StringList(lines), 0, node)
-        # result[1] is the content node,
-        # result[1][0] is the signature line
-        # result[1][1] is the fieldlist and description
-        # result[1][1][0] is the fieldlist
-        if (result[1][1].children and
-            isinstance(description[0], nodes.field_list)):
-            cresult = description.deepcopy()
-            target = description
-            target.clear()
-            target.append(cresult[0])
-            target.extend(node)
-            target.extend(cresult[1:])
-        else:
-            cresult = description.deepcopy()
-            target = description
-            target.clear()
-            target.extend(node)
-            target.extend(cresult)
-        return result
 
-    def run_add_specializers(self, result):
-        package = self.env.temp_data.get('cl:package')
-        name = self.names[0][1]
-        description = result[1][-1]
-        specializers = METHODS[package].get(name, {}).keys()
-        if specializers:
-            description.append(nodes.paragraph(text="Specializes"))
-            spec = nodes.bullet_list()
-            spec += [specializer(s, self.state) for s in specializers]
-            description.children.append(spec)
+        # Insert parsed nodes just after the field list.
+        content = get_content_node(result)
+        index = fieldlist_index(content)
+        if index is not None:
+            index += 1
+            start = content[:index]
+        else:
+            start = []
+            index = 0
+        rest = content[index:]
+        content.children = start + node + rest
         return result
 
     def run(self):
         result = super(CLsExp, self).run()
         if "nodoc" not in self.options:
             self.run_add_doc(result)
-        if self.objtype == "generic" and "nospecializers" not in self.options:
-            self.run_add_specializers(result)
         return result
 
     def cl_doc_string(self, objtype=None):
@@ -439,55 +492,145 @@ class CLsExp(ObjectDescription):
         symbol can't be found.
         """
         package = self.env.temp_data.get('cl:package')
-        name = self.names[0][1]
+        name = self.cl_symbol_name()
         objtype = objtype or self.objtype
-
         possible_strings = DOC_STRINGS[package][name]
 
-        # XXX This isn't the best, the objtype is generic but the
-        # docstring will be under genericFunction because of the JSON
-        # encoder and changing the directive name doesn't seem to help
-        # either.
-        if objtype == "generic":
-            objtype = "genericFunction"
         string = possible_strings.get(objtype, "")
         return string
 
 
+class CLGeneric(CLsExp):
+
+    option_spec = {
+        'nodoc': bool_option,
+        'noindex': bool_option,
+        'nospecializers': bool_option,
+    }
+
+    def run_add_specializers(self, result):
+        package = self.env.temp_data.get('cl:package')
+        name = self.cl_symbol_name()
+        description = get_content_node(result)
+        specializers = METHODS[package].get(name, {}).keys()
+        if specializers:
+            description.append(nodes.paragraph(text="Specializes:"))
+            spec = nodes.bullet_list()
+            spec += [specializer_xref(package + ":" + name, s, self.state)
+                     for s in specializers]
+            description.append(spec)
+        return result
+
+    def run(self):
+        result = super(CLGeneric, self).run()
+        if "nospecializers" not in self.options:
+            self.run_add_specializers(result)
+        return result
+
+
 class CLMethod(CLsExp):
 
     option_spec = {
         'nodoc': bool_option,
         'noindex': bool_option,
         'noinherit': bool_option,
+        'nospecializers': bool_option,
     }
 
+    doc_field_types = [
+        Field('specializer', label=l_('Specializer'), has_arg=False,
+              names=('specializer')),
+        GroupedField('parameter', label=l_('Parameters'),
+                     names=('param', 'parameter', 'arg', 'argument',
+                            'keyword', 'kwparam')),
+        Field('returnvalue', label=l_('Returns'), has_arg=False,
+              names=('returns', 'return')),
+    ]
+
+    def get_index_name(self, name, type):
+        package = self.env.temp_data.get('cl:package')
+        specializer = self.arguments
+        spec_args = qualify_sexp(package, specializer[0].split(" ")[1:])
+        specializer = " ".join(spec_args)
+        return type + ":" + name + "(" + specializer.lower() + ")"
+
+    def get_index_text(self, name, type):
+        specializer = self.arguments
+        spec_args = specializer[0].split(" ")[1:]
+        specializer = " ".join(spec_args)
+        return _('%s (%s) (Lisp %s)') % (name.lower().split(":")[-1],
+                                         specializer.lower(), type)
+
+    def add_target_and_index(self, name, sig, signode):
+        # node target
+        type, name = name
+
+        if 'cl:package' in self.env.temp_data:
+            package = self.options.get(
+                'module', self.env.temp_data.get('cl:package'))
+            name = package.lower() + ":" + name
+        else:
+            return
+
+        indexname = self.get_index_name(name, type)
+        if name not in self.state.document.ids:
+            signode['names'].append(name)
+            signode['ids'].append(indexname)
+            signode['first'] = (not self.names)
+            self.state.document.note_explicit_target(signode)
+            inv = self.env.domaindata['cl']['methods']
+            # TODO (RS) reenable this checking based on doc and type.
+            # if name in inv:
+            #     self.state_machine.reporter.warning(
+            #         'duplicate symbol description of %s, ' % name +
+            #         'other instance in ' + self.env.doc2path(inv[name][0]),
+            #         line=self.lineno)
+            sig = " ".join(qualify_sexp(package.lower(), sig.split(" ")[1:]))  # trim method name
+            if name in inv:
+                inv[name][sig] = (self.env.docname, self.objtype)
+            else:
+                inv[name] = {sig: (self.env.docname, self.objtype)}
+
+        indextext = self.get_index_text(name, type)
+        if indextext:
+            self.indexnode['entries'].append(('single', indextext, indexname, ''))
+
     def cl_doc_string(self):
         """
         Resolve a symbols doc string. Will raise KeyError if the
         symbol can't be found.
         """
         package = self.env.temp_data.get('cl:package')
-        name = self.names[0][1]
-        objtype = self.objtype
+        name = self.cl_symbol_name()
 
-        if objtype == "method":
-            specializer = self.arguments
-            spec = specializer[0].split(" ")[1:]
-            method_doc = METHODS[package].get(name, {})
-            key = tuple([parse_specializer_symbol(sym, package)
-                         for sym in spec])
-            if key not in method_doc:
-                self.state_machine.reporter.warning("Can't find method %s:%s specializer %s, available specializers are %s" %
-                                               (package, name, key, method_doc.keys()))
-            doc = method_doc.get(key, "")
-            if doc:
-                return doc
+        specializer = self.arguments
+        spec = specializer[0].split(" ")[1:]
+        method_doc = METHODS[package].get(name, {})
+        key = tuple([parse_specializer_symbol(sym, package)
+                     for sym in spec])
+        if key not in method_doc:
+            self.state_machine.reporter.warning("Can't find method %s:%s specializer %s, available specializers are %s" %
+                                           (package, name, key, method_doc.keys()))
+        doc = method_doc.get(key, "")
+        if doc:
+            return doc
 
         if "noinherit" not in self.options:
             return super(CLMethod, self).cl_doc_string("generic")
         return ""
 
+    def run(self):
+        result = super(CLMethod, self).run()
+        field_list = self.get_field_list(result)
+
+        spec = specializer(self.arguments[0].split()[1:], self.state)
+        field_list.append(
+            nodes.field('',
+                        nodes.field_name('', "Specializer"),
+                        nodes.field_body('',
+                                         nodes.paragraph('', '', spec))))
+        return result
+
 
 class CLCurrentPackage(Directive):
     """
@@ -540,7 +683,7 @@ class CLDomain(Domain):
     directives = {
         'package': CLCurrentPackage,
         'function': CLsExp,
-        'generic': CLsExp,
+        'generic': CLGeneric,
         'macro': CLsExp,
         'variable': CLsExp,
         'type': CLsExp,
@@ -549,22 +692,25 @@ class CLDomain(Domain):
 
     roles = {
         'symbol': CLXRefRole(),
+        'method': CLXRefRole(),
     }
     initial_data = {
         'symbols': {},
+        'methods': {},
     }
 
     def clear_doc(self, docname):
-        for fullname, (fn, _) in self.data['symbols'].items():
-            if fn == docname:
-                del self.data['symbols'][fullname]
+        for fullname, docs in self.data['symbols'].items():
+            for (fn, _) in docs:
+                if fn == docname:
+                    del self.data['symbols'][fullname]
 
     def find_obj(self, env, name):
         """Find a Lisp symbol for "name", perhaps using the given package
         Returns a list of (name, object entry) tuples.
         """
         symbols = self.data['symbols']
-        matches = []
+        name = name.lower()
         if ":" in name:
             if name in symbols:
                 return [(name, symbols[name])]
@@ -580,9 +726,30 @@ class CLDomain(Domain):
                 return False
             return filter(filter_symbols, symbols.items())
 
+    def find_method(self, env, name, node):
+        """Find a Lisp symbol for "name", perhaps using the given package
+        Returns a list of (name, object entry) tuples.
+        """
+        methods = self.data['methods']
+        name = name.lower()
+        sexp = name.split(" ")
+        generic = sexp[0]
+        specializer = " ".join(sexp[1:])
+        if generic in methods:
+            if specializer in methods[generic]:
+                return [methods[generic][specializer]]
+            else:
+                env.warn_node('can\'t find method %s' % (name), node)
+        else:
+            env.warn_node('can\'t find generic %s' % (name), node)
+
     def resolve_xref(self, env, fromdocname, builder,
                      typ, target, node, contnode):
-        matches = self.find_obj(env, target.upper())
+        if " " in target:
+            matches = self.find_method(env, target.upper(), node)
+        else:
+            matches = self.find_obj(env, target.upper())
+
         if not matches:
             return None
         elif len(matches) > 1:
@@ -590,13 +757,28 @@ class CLDomain(Domain):
                 'more than one target found for cross-reference '
                 '%r: %s' % (target, ', '.join(match[0] for match in matches)),
                 node)
-        name, obj = matches[0]
-        return make_refnode(builder, fromdocname, obj[0], name,
-                            contnode, name)
+        # TODO (RS) this just chooses the first symbol, instead every
+        # symbol should be presented.
+
+        if " " in target:
+            sexp = target.split(" ")
+            generic = sexp[0].lower()
+            specializer = " ".join(sexp[1:])
+            name = generic
+            filename = matches[0][0]  # the first filename
+            link = "method" + ":" + generic + "(" + specializer + ")"
+        else:
+            name = matches[0][0]  # the symbol name
+            filename = matches[0][1][0][0]  # the first filename
+            type = matches[0][1][0][1]  # the first type
+            link = type + ":" + name
+        return make_refnode(builder, fromdocname, filename,
+                            link, contnode, name)
 
     def get_symbols(self):
-        for refname, (docname, type) in self.data['symbols'].iteritems():
-            yield (refname, refname, type, docname, refname, 1)
+        for refname, docs in self.data['symbols'].iteritems():
+            for (docname, type) in docs:
+                yield (refname, refname, type, docname, refname, 1)
 
 
 def code_regions(text):
@@ -657,8 +839,17 @@ def index_package(package, package_path, quicklisp, lisps):
         for type in ALL_TYPES:
             if not type in v:
                 continue
+            # XXX This isn't the best, the objtype is generic but the
+            # docstring will be under genericFunction because of the JSON
+            # encoder and changing the directive name doesn't seem to help
+            # either.
+            if type == "genericFunction":
+                cl_type = "generic"
+            else:
+                cl_type = type
+
             # enable symbol references for symbols
-            DOC_STRINGS[package][k][type] = code_regions(v[type])
+            DOC_STRINGS[package][k][cl_type] = code_regions(v[type])
 
         # extract methods
         if "methods" in v:
